import json
import logging
import os
import re
import unittest

from pararamio import Pararamio


class SensitiveFormatter(logging.Formatter):
    """Formatter that removes sensitive information in urls."""

    @staticmethod
    def _filter(s):
        return re.sub(r'password":\s*\"\w+\"', r'password": "*******"', s)

    def format(self, record):
        original = logging.Formatter.format(self, record)
        return self._filter(original)


class BasePararamioTest(unittest.TestCase):
    def setUp(self):
        self.user1 = json.loads(os.environ.get('PARARAMIO_FIRST_USER', '{}'))
        self.user2 = json.loads(os.environ.get('PARARAMIO_SECOND_USER', '{}'))
        self.client = Pararamio(**self.user1, cookie_path='test.cookie')
        self.client2 = Pararamio(**self.user2, cookie_path='test.cookie2')
        self.__post = None
        if bool(os.environ.get('DEBUG', False)):
            logger = logging.getLogger('pararamio')
            logger.setLevel(logging.DEBUG)
            ch = logging.StreamHandler()
            ch.setLevel(logging.DEBUG)
            ch.setFormatter(SensitiveFormatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
            logger.addHandler(ch)



from datetime import datetime
from typing import Any, Dict, Iterable, List, Optional, TYPE_CHECKING, Tuple, Union, cast
from urllib.parse import quote_plus

import pararamio
from pararamio._types import QuoteRangeT
from pararamio.constants import POSTS_LIMIT
from pararamio.exceptions import PararamioLimitExceededException, PararamioMethodNotAllowed, PararamioRequestException, PararamioValidationException
from pararamio.file import File
from pararamio.post import Post
from pararamio.utils.helpers import encode_chats_ids, format_datetime, join_ids, parse_datetime


if TYPE_CHECKING:
    from pararamio.client import Pararamio
    from pararamio.user import User

__all__ = ('Chat',)

ATTR_FORMATTERS = {
    'time_edited':      parse_datetime,
    'time_created':     parse_datetime,
    'user_time_edited': parse_datetime,
}


def check_result(result: dict) -> bool:
    return 'chat_id' in result


class Chat:
    id: int
    title: str
    history_mode: str
    description: Optional[str]
    posts_count: int
    pm: bool
    e2e: bool
    time_created: datetime
    time_updated: datetime
    time_edited: Optional[datetime]
    author_id: int
    two_step_required: bool
    org_visible: bool
    organization_id: Optional[int]
    posts_live_time: Optional[int]
    allow_api: bool
    read_only: bool
    tnew: bool
    adm_flag: bool
    custom_title: Optional[str]
    is_favorite: bool
    inviter_id: Optional[int]
    tshow: bool
    user_time_edited: datetime
    history_start: int
    pinned: List[int]
    thread_groups: List[int]
    thread_users: List[int]
    thread_admins: List[int]
    thread_users_all: List[int]
    last_msg_author_id: Optional[int]
    last_msg_author: str
    last_msg_bot_name: str
    last_msg_text: str
    last_msg: str
    last_read_post_no: int
    thread_guests: List[int]
    _data: Dict[str, Any]
    _client: 'Pararamio'
    load_on_key_error: bool

    def __init__(self, client: 'Pararamio', id: int = None, load_on_key_error: bool = True, **kwargs):
        if id is None:
            id = kwargs.get('chat_id', None)
            if id is None:
                id = kwargs['thread_id']
        self.id = int(id)
        self._data = {}
        if kwargs:
            self._data = {**kwargs, 'id': id}
        self.load_on_key_error = load_on_key_error
        self._client = client

    def __getattr__(self, name: str):
        fmt_fnc = ATTR_FORMATTERS.get(name, None)
        if fmt_fnc:
            return fmt_fnc(self._data, name)
        try:
            return self._data[name]
        except KeyError:
            if self.load_on_key_error:
                self.load()
                return self._data[name]
            raise

    def __str__(self):
        title = self._data.get('title', '')
        id_ = self.id or ''
        return f'{id_} - {title}'

    def __eq__(self, other: object):
        if not isinstance(other, Chat):
            return id(other) == id(self)
        return self.id == other.id

    def __contains__(self, item: Union[Post, 'User']) -> bool:
        if isinstance(item, Post):
            return item.chat == self
        if isinstance(item, pararamio.user.User):
            return item.id in self.thread_users_all
        return False

    @property
    def client(self) -> 'Pararamio':
        return self._client

    def load(self) -> 'Chat':
        if self.id is None:
            raise PararamioMethodNotAllowed(f'Load is not allow for new {self.__class__.__name__}')
        chats = self.load_chats(self._client, [self.id])
        if len(chats) != 1:
            raise PararamioRequestException(f'failed to load data for chat id {self.id}')
        self._data = chats[0]._data
        return self

    def edit(self, **kwargs) -> None:
        """
        :param kwargs:
        available args
        {
            "title": "My title",
            "description": "Description",
            "posts_live_time": null,  // null,int,float sec
            "two_step_required": false,
            "history_mode": "all",
            "cover_path": null,
            "org_visible": false,
            "allow_api": true,
            "read_only": false
        }
        :return:
        """
        url = f'/core/chat/{self.id}'
        check_result(self._client.api_put(url, data=kwargs))

    def transfer(self, org_id: int) -> bool:
        url = f'/core/chat/{self.id}/transfer/{org_id}'
        return check_result(self._client.api_post(url, {}))

    def delete(self):
        url = f'/core/chat/{self.id}'
        return check_result(self._client.api_delete(url))

    def hide(self) -> bool:
        url = f'/core/chat/{self.id}/hide'
        return check_result(self._client.api_post(url))

    def show(self) -> bool:
        url = f'/core/chat/{self.id}/show'
        return check_result(self._client.api_post(url))

    def favorite(self) -> bool:
        url = f'/core/chat/{self.id}/favorite'
        return check_result(self._client.api_post(url))

    def unfavorite(self) -> bool:
        url = f'/core/chat/{self.id}/unfavorite'
        return check_result(self._client.api_post(url))

    def enter(self) -> bool:
        url = f'/core/chat/{self.id}/enter'
        return check_result(self._client.api_post(url))

    def quit(self) -> bool:
        url = f'/core/chat/{self.id}/quit'
        return check_result(self._client.api_post(url))

    def set_custom_title(self, title: str) -> bool:
        url = f'/core/chat/{self.id}/custom_title'
        return check_result(self._client.api_post(url, {'title': title}))

    def sync(self) -> None:
        # url = '/core/chat/sync'
        # return self.client.api_post(url, {
        #    'ids':       encode_chat_id(self.id, self._data.get('posts_count', 0), self._data.get('last_read_post_no', 0)),
        #    'sync_time': self._data.get('time_updated', datetime.now().strftime(DATETIME_FORMAT))
        # })
        # TODO: Implement sync
        raise NotImplementedError()

    def add_users(self, ids: List[int]) -> bool:
        url = f'/core/chat/{self.id}/user/{join_ids(ids)}'
        return check_result(self._client.api_post(url))

    def delete_users(self, ids: List[int]) -> bool:
        url = f'/core/chat/{self.id}/user/{join_ids(ids)}'
        return check_result(self._client.api_delete(url))

    def add_admins(self, ids: List[int]) -> bool:
        url = f'/core/chat/{self.id}/admin/{join_ids(ids)}'
        return check_result(self._client.api_post(url))

    def delete_admins(self, ids: List[int]) -> bool:
        url = f'/core/chat/{self.id}/admin/{join_ids(ids)}'
        return check_result(self._client.api_delete(url))

    def add_groups(self, ids: List[int]) -> bool:
        url = f'/core/chat/{self.id}/group/{join_ids(ids)}'
        return check_result(self._client.api_post(url))

    def delete_groups(self, ids: List[int]) -> bool:
        url = f'/core/chat/{self.id}/group/{join_ids(ids)}'
        return check_result(self._client.api_delete(url))

    def _load_posts(self, start_post_no: int = -50, end_post_no: int = -1) -> List['Post']:
        url = f'/msg/post?chat_id={self.id}&range={start_post_no}x{end_post_no}'
        if (start_post_no < 0 <= end_post_no) or (start_post_no >= 0 > end_post_no):
            raise PararamioValidationException('start_post_no and end_post_no can only be negative or positive at the same time')
        if 0 > start_post_no > end_post_no:
            raise PararamioValidationException('range start_post_no must be greater then end_post_no')
        if 0 <= start_post_no > end_post_no:
            raise PararamioValidationException('range start_post_no must be smaller then end_post_no')
        _absolute = abs(end_post_no - start_post_no)
        if start_post_no < 0:
            _absolute = + 1
        if _absolute >= POSTS_LIMIT:
            raise PararamioLimitExceededException('max post load limit is {}'.format(POSTS_LIMIT - 1))
        res = self._client.api_get(url).get('posts', [])
        if not res:
            return []
        return [Post(chat=self, **post) for post in res]

    def _lazy_posts_loader(self, start_post_no: int = -50, end_post_no: int = -1) -> Iterable['Post']:
        if (start_post_no < 0 <= end_post_no) or (start_post_no >= 0 > end_post_no):
            raise PararamioValidationException('start_post_no and end_post_no can only be negative or positive at the same time')
        if 0 > start_post_no > end_post_no:
            raise PararamioValidationException('range start_post_no must be greater then end_post_no')
        if 0 <= start_post_no > end_post_no:
            raise PararamioValidationException('range start_post_no must be smaller then end_post_no')
        absolute = abs(end_post_no - start_post_no)
        start, end = start_post_no, end_post_no
        if absolute > POSTS_LIMIT:
            end = start_post_no + POSTS_LIMIT - 1
        posts = iter(self._load_posts(start, end))
        counter = 0
        for post_no in range(start_post_no, end_post_no):
            try:
                yield next(posts)
            except StopIteration:
                counter += 1
                res = self._load_posts(start + POSTS_LIMIT * counter, end + POSTS_LIMIT * counter)
                if not res:
                    return
                posts = iter(res)

    def posts(self, start_post_no: int = -50, end_post_no: int = -1) -> List['Post']:
        return self._load_posts(start_post_no=start_post_no, end_post_no=end_post_no)

    def read_status(self, post_no: int) -> bool:
        url = f'/msg/lastread/{self.id}'
        res = self._client.api_post(url, {'post_no': post_no})
        if not res:
            return False
        if 'post_no' in res:
            self._data['last_read_post_no'] = res['post_no']
        if 'posts_count' in res:
            self._data['posts_count'] = res['posts_count']
        return True

    def post(self, text: str, quote_range: QuoteRangeT = None, reply_no: int = None, file: File = None) -> 'Post':
        if self.id is None:
            raise ValueError('can not post file to new chat')
        return Post.create(self, text=text, reply_no=reply_no, quote=cast(str, quote_range['text']) if quote_range else None)

    def direct_upload_file(self, file_path: str, filename: str = None, reply_no: int = None, quote_range: str = None) -> File:
        if self.id is None:
            raise ValueError('can not upload file to new chat')
        return self._client.direct_upload_file(file_path=file_path, chat_id=self.id, filename=filename, reply_no=reply_no, quote_range=quote_range)

    def upload_file(self, file_path: str, filename: str = None) -> File:
        if self.id is None:
            raise ValueError('can not upload file to new chat')
        return self._client.upload_file(file_path=file_path, chat_id=self.id, filename=filename)

    @classmethod
    def load_chats(cls, client: 'Pararamio', ids: List[int]) -> List['Chat']:
        url = f'/core/chat?ids={join_ids(ids)}'
        res = client.api_get(url)
        if res and 'chats' in res:
            return [cls(client, **data) for data in client.api_get(url).get('chats', [])]
        raise PararamioRequestException('failed to load data for chats ids: {}'.format(','.join(map(str, ids))))

    @classmethod
    def post_search(cls,
                    client: 'Pararamio',
                    q: str,
                    order_type: str = 'time',
                    page: int = 1,
                    chat_id: int = None,
                    limit: Optional[int] = POSTS_LIMIT) -> Tuple[int, Iterable['Post']]:
        if not limit:
            limit = POSTS_LIMIT
        url = f'/posts/search?q={quote_plus(q)}&order_type={order_type}&page={page}&limit={limit}'
        if chat_id is not None:
            url += f'&th_id={chat_id}'

        res = client.api_get(url)
        if 'posts' not in res:
            raise PararamioRequestException('failed to perform search')
        created_chats = {}

        def create_post(data):
            nonlocal created_chats
            _chat_id = data['thread_id']
            post_no = data['post_no']
            if _chat_id not in created_chats:
                created_chats[_chat_id] = cls(client, id=_chat_id)
            return Post(created_chats[_chat_id], post_no=post_no)

        return res['count'], map(create_post, res['posts'])

    @classmethod
    def create(cls, client: 'Pararamio', title: str, description: str = '', users: List[int] = None, groups: List[int] = None, **kwargs) -> 'Chat':
        """
        Create new chat in pararam
        :param client:
        :param title:
        :param description:
        :param users:
        :param groups:
        :param kwargs: available
        {
            "organization_id": null,  // null,int
            "posts_live_time": null,  // null,int,float sec
            "two_step_required": false,
            "cover_path": null,
            "history_mode": "all",
            "org_visible": false,
            "allow_api": true,
            "read_only": false,
        }
        :return:
        """
        if users is None:
            users = []
        if groups is None:
            groups = []
        data = {'title': title, 'description': description, 'users': users, 'groups': groups, **kwargs}

        res = client.api_post('/core/chat', data)
        id_: int = res['chat_id']
        return cls(client, id_)

    @classmethod
    def create_private_chat(cls, client: 'Pararamio', user_id: int) -> 'Chat':
        url = f'/core/chat/pm/{user_id}'
        res = client.api_post(url)
        id_: int = res['chat_id']
        return cls(client, id=id_)

    @staticmethod
    def sync_chats(client: 'Pararamio', chats_ids: List[Tuple[int, int, int]], sync_time: datetime = None) -> Dict[str, Any]:
        """

        :param client:
        :param chats_ids: List[Tuple[chat_id,posts_count,last_read_post_no]
        :param sync_time: updates from this time
        :return: {
            "new": int[],   // список чатов с tshow=True, но не преданных с клиента
            "updated": int[],  // список чатов с tshow=True и time_edited > sync_time
            "removed": int[], // список чатов переданных с клиента, но не найденных с фильтром tshow=True
            "status": {  // тут будут статусы чатов, если они не совпадают с переданными
                123: {  // chat_id
                    "posts_count": 47,
                    "last_read_post_no": 45,
                    "last_msg": "vlasov: фыв",
                    "time_updated": "2018-08-23T09:18:56.143850Z"
                },
               ...
            }

        """
        url = '/core/chat/sync'
        data = {
            'ids': encode_chats_ids(chats_ids)
        }
        if sync_time:
            data['sync_time'] = format_datetime(sync_time)
        return client.api_post(url)



from datetime import datetime, timezone
from typing import List, Optional

from pararamio._types import QuoteRangeT
from pararamio.client import Pararamio
from pararamio.utils.helpers import format_datetime, parse_datetime


__all__ = ('DeferredPost',)
ATTR_FORMATTERS = {
    'time_created': parse_datetime,
    'time_sending': parse_datetime
}


class DeferredPost:
    """
    {
            "id": 1,
            "user_id": 1,
            "chat_id": 27,
            "data": {  // данные для обработки, дублируются
                "user_id": 1,
                "chat_id": 27,
                "text": "deferred post",
                "reply_no": null,
                "quote_range": null
            },
            "time_created": "2019-10-16T11:09:39.011395Z",
            "time_sending": "2019-10-17T00:00:00Z"
        }
    """
    id: int
    user_id: int
    chat_id: int
    text: str
    reply_no: Optional[int]
    time_created: datetime
    time_sending: datetime
    _data: dict
    _client: 'Pararamio'

    def __init__(self, client: 'Pararamio', id: int, **kwargs):
        self._client = client
        self.id = id
        self._data = {'id': id, **kwargs}

    def __getattr__(self, name):
        fmt_fnc = ATTR_FORMATTERS.get(name, None)
        if fmt_fnc:
            return fmt_fnc(self._data, name)
        if name in self._data.get('data', {}):
            return self._data['data'][name]
        return self._data[name]

    def __str__(self):
        text = self._data.get('text', None)
        if text is None:
            self.load()
            text = self._data['text']
        return text

    def delete(self):
        url = f'/msg/deferred/{self.id}'
        self._client.api_delete(url)

    @classmethod
    def create(cls, client: 'Pararamio', chat_id: int, text: str, time_sending: datetime, reply_no: int = None,
               quote_range: QuoteRangeT = None) -> 'DeferredPost':
        url = '/msg/deferred'
        data = {
            'chat_id':      chat_id,
            'text':         text,
            'time_sending': format_datetime(time_sending),
            'reply_no':     reply_no,
            'quote_range':  quote_range
        }
        res = client.api_post(url, data)
        return cls(client, id=int(res['deferred_post_id']), **{
            'chat_id':      chat_id,
            'data':         data,
            "time_sending": data['time_sending']
        })

    @classmethod
    def get_deferred_posts(cls, client: 'Pararamio') -> List['DeferredPost']:
        url = '/msg/deferred'
        res = client.api_get(url).get('posts', [])
        return [cls(client, **post) for post in res]



import os
import tempfile
import time
import unittest
from datetime import datetime, timedelta, timezone
from http.cookiejar import MozillaCookieJar
from unittest.mock import MagicMock, patch
import logging

from pararamio.chat import Chat
from pararamio.client import Pararamio
from pararamio.deferred_post import DeferredPost
from pararamio.exceptions import PararamioException, PararamioValidationException
from pararamio.post import Post
from pararamio.user import User
from tests.integrations._base import BasePararamioTest


RESOURCES_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), '../resources')


class PararamioClientTest(BasePararamioTest):
    def setUp(self):
        super().setUp()
        if int(os.environ.get('DEBUG', 0)):
            logging.basicConfig(level=logging.DEBUG)
        self.recipient_email = os.environ.get('PARARAMIO_RECIPIENT_EMAIL', '')
        self.recipient_chat_id = os.environ.get('PARARAMIO_RECIPIENT_CHAT_ID', '')

    def test_001_login_without_cookie(self):
        client = Pararamio(**self.user1)
        self.assertIsNotNone(client.profile)

    def test_002_login_with_cookie(self):
        with tempfile.NamedTemporaryFile(suffix='.cookie') as tmp:
            MozillaCookieJar(tmp.name).save()
            client = Pararamio(**self.user1, cookie_path=tmp.name)
            self.assertIsNotNone(client.profile, 'Authentication failed')
            self.assertTrue(os.path.exists(tmp.name), 'Cookie path is not found')
            del client
            client = Pararamio(login='', password='', key='', cookie_path=tmp.name)
            self.assertIsNotNone(client.profile, 'Cookie authentication failed')

    def test_003_login_with_broken_cookie(self):
        with tempfile.NamedTemporaryFile(suffix='.cookie') as tmp:
            tmp.write('broken_cookie'.encode())
            client = Pararamio(**self.user1, cookie_path=tmp.name, ignore_broken_cookie=True)
            self.assertIsNotNone(client.profile, 'Authentication failed')
            self.assertTrue(os.path.exists(tmp.name), 'Cookie path is not found')

    def test_004_profile(self):
        self.assertEqual(self.client.profile.get('email'), self.user1['login'])

    def test_005_send_delete_file(self):
        d = {
            'size':     23050,
            'type':     'image/png',
            'filename': 'test.png'
        }
        post = None
        chat = Chat(self.client, self.recipient_chat_id)
        file = chat.upload_file(file_path=os.path.join(RESOURCES_PATH, 'test.png'))

        for _post in chat.posts(-5, -1):
            _post.load()
            if _post.file and _post.file.guid == file.guid:
                post = _post

        self.assertEqual(post.file.filename, d['filename'])
        self.assertIsNotNone(file)
        file.delete()
        self.assertDictEqual({
            'size':     file.size,
            'type':     post.file._data['mime_type'],
            'filename': file.filename
        },
            d)

    def test_006_list_own_threads(self):
        chats = self.client.list_chats()
        self.assertIsNotNone(chats)
        chat = next(chats)
        chat_ = Chat(self.client, chat.id)
        chat_.load()
        self.assertTrue(chat == chat_)

    def test_007_list_posts(self):
        chat = Chat(self.client, self.recipient_chat_id)
        self.assertTrue(chat.posts(-10, -1))
        self.assertTrue(chat.posts(0, 1))
        chat.load()
        exception = False
        try:
            chat.posts(1, 0)
        except PararamioValidationException:
            exception = True
        self.assertTrue(exception, 'posts load with invalid start end position')
        exception = False
        try:
            chat.posts(-1, -10)
        except PararamioValidationException:
            exception = True
        self.assertTrue(exception, 'posts load with invalid start end position')

    def test_008_search_user(self):
        self.assertIsNotNone(self.client.search_user('tamtambottester2@gmail.com'))
        self.assertIsNotNone(self.client.search_user('tamtambottester2'))
        self.assertIsNotNone(self.client.search_user('tamtam_bot_tester2'))

    def test_009_reply_to_post(self):
        chat = Chat(self.client, self.recipient_chat_id)
        post = chat.post('test')
        reply = post.reply('ReplyTest')
        self.assertEqual(post.post_no, reply.reply_no)

    def test_010_search_group(self):
        self.assertIsNotNone(self.client.search_group(''))

    def test_011_list_users(self):
        u = User(self.client, 58)
        u.load()

        self.assertTrue(u.unique_name)

    def test_012_chat_load(self, ):
        u = self.client.search_user('tamtambottester2@gmail.com')[0]
        chat = u.get_pm_thread()
        chat.load()
        pc = chat.posts_count
        u.post('text')
        chat.load()
        self.assertTrue(chat.posts_count == pc + 1)

    def test_013_read_status(self):
        u = self.client.search_user(self.recipient_email)[0]
        chat = u.get_pm_thread()
        chat.load()
        post = chat.post('test')
        chat.read_status(post.post_no + 1)
        self.assertTrue(chat.last_read_post_no == post.in_thread_no)

    def test_014_edit(self):
        u = self.client.search_user(self.recipient_email)[0]
        post = u.post('test')
        post.edit(text='TestTestStringString')
        chat = u.get_pm_thread()
        np = Post(chat, post.post_no)
        np.load()
        self.assertEqual('TestTestStringString', np.text)

    def test_015_send_and_delete_private_message(self):
        u = self.client.search_user(self.recipient_email)[0]
        post = u.post('test')
        self.assertIsNotNone(post)
        self.assertIsNotNone(post.delete())

    def test_016_send_and_delete_thread_message(self):
        chat = Chat(self.client, self.recipient_chat_id)
        post = chat.post('test')
        self.assertIsNotNone(post)
        self.assertIsNotNone(post.delete())

    def test_017_chat_methods(self):
        d = {'title': 'Test Create Chat', 'users': [self.client.profile['id']], 'pm': False, 'member_ids': [self.client.profile['id']]}
        chat = Chat.create(self.client2, **d)
        self.assertTrue(User(self.client, self.client.profile['id']) in chat)
        self.assertFalse(User(self.client, -1) in chat)
        self.assertTrue(Post(chat, post_no=1, ) in chat)
        self.assertFalse(Post(Chat(self.client, id=100), post_no=1) in chat)
        self.assertTrue(chat.id in [ch.id for ch in self.client.list_chats()], 'Thread is not created')
        chat.add_admins([self.client.profile['id']])
        chat.load()
        chat2 = Chat(self.client, chat.id).load()
        self.assertTrue(chat2.adm_flag, 'Admin is not set')
        time.sleep(5)
        chat.delete_admins([self.client.profile['id']])
        chat2.load()
        self.assertFalse(chat2.adm_flag, 'User is still admin')
        chat.delete()
        self.assertFalse(chat.id in [tr.id for tr in self.client.list_chats()], 'Thread is still exists')

    def test_018_deferred_posts(self):
        time_sending = datetime.now(tz=timezone.utc) + timedelta(days=1)
        text = f'{time_sending.timestamp()}: test deferred posts !'
        result = DeferredPost.create(self.client, self.recipient_chat_id, text, time_sending)
        self.assertTrue(result)
        posts = DeferredPost.get_deferred_posts(self.client)
        post = [p for p in posts if p.text == text]
        self.assertTrue(bool(post), 'Post is empty')
        self.assertTrue(post[0].text == text, 'text is not match')
        self.assertTrue(int(post[0].time_sending.timestamp()) == int(time_sending.timestamp()))
        post[0].delete()

    def test_019_posts_tree(self):
        chat = Chat(self.client, self.recipient_chat_id, posts_count=100)
        with patch.object(Post, 'rerere', return_value=[
            Post(chat, post_no=96, text='post no 96', reply_no=None),
            Post(chat, post_no=99, text='post no 99', reply_no=96),
            Post(chat, post_no=100, text='post no 100', reply_no=99, meta={})
        ]):
            with patch.object(chat, '_lazy_posts_loader', return_value=iter([
                Post(chat, post_no=96, text='post no 96', reply_no=None),
                Post(chat, post_no=97, text='post no 97', reply_no=None),
                Post(chat, post_no=98, text='post no 98', reply_no=96),
                Post(chat, post_no=99, text='post no 99', reply_no=96),
            ])):
                post = Post(chat, post_no=100, text='post no 100', reply_no=99, meta={})
                self.assertListEqual([96, 98, 99, 100], list(post.get_tree().keys()))
        many_fake_posts = [
            Post(chat, post_no=1, text='post no 1', reply_no=None),
            *[Post(chat, text=f'post no {i}', post_no=i, reply_no=i - 1) for i in range(2, 2000)],
            Post(chat, post_no=2000, text='post no 2000', reply_no=1999, meta={})
        ]

        def _load_posts_side_effect(start, end):
            return iter(many_fake_posts[start: end])

        with patch.object(Post, 'rerere', return_value=[many_fake_posts[0], *many_fake_posts[1000:]]):
            chat._load_posts = MagicMock(side_effect=_load_posts_side_effect)
            posts = list(many_fake_posts[-1].get_tree().keys())
            self.assertListEqual([many_fake_posts[0].post_no, *[p.post_no for p in many_fake_posts[1000:]]], posts)

        with patch.object(Post, 'rerere', return_value=[]):
            chat._load_posts = MagicMock(side_effect=lambda a, b: [])
            self.assertListEqual([2000], list(many_fake_posts[-1].get_tree().keys()))

    def test_020_send_and_delete_message_by_email(self):
        text = 'test by mail'
        try:
            self.client.post_private_message_by_user_email(self.recipient_email + '1111', text)
            self.assertTrue(False, 'the message cannot be sent to a non-existent email')
        except PararamioException:
            pass
        post = self.client.post_private_message_by_user_email(self.recipient_email, text)
        self.assertIsNotNone(post)
        self.assertTrue(post.text == text)
        self.assertIsNotNone(post.delete())

    def test_021_send_and_delete_message_by_id(self):
        text = 'test by mail'
        try:
            self.client.post_private_message_by_user_email(-1, text)
            self.assertTrue(False, 'the message cannot be sent to a non-existent id')
        except PararamioException:
            pass
        post = self.client.post_private_message_by_user_id(self.client2.profile['id'], text)
        self.assertIsNotNone(post)
        self.assertTrue(post.text == text)
        self.assertIsNotNone(post.delete())

    @classmethod
    def tearDownClass(cls):
        if os.path.exists('test.cookie'):
            os.remove('test.cookie')
        if os.path.exists('test.cookie2'):
            os.remove('test.cookie2')


def suite():
    return unittest.TestLoader().loadTestsFromTestCase(PararamioClientTest)


if __name__ == "__main__":
    unittest.main()



__all__ = (
    'PararamioException',
    'PararamioRequestException',
    'PararamioMethodNotAllowed',
    'PararamioXSFRRequestError',
    'PararamioServerResponseException',
    'PararamioAuthenticationException',
    'PararamioValidationException',
    'PararamioHTTPRequestException',
    'PararamioLimitExceededException',
)

import json
from json import JSONDecodeError
from typing import IO, List, Optional, Union
from urllib.error import HTTPError


class PararamioException(Exception):
    pass


class PararamioValidationException(PararamioException):
    pass


class PararamioHTTPRequestException(HTTPError, PararamioException):
    _response: Union[bytes, None]
    fp: Optional[IO[bytes]]

    def __init__(self, url: str, code: int, msg: str, hdrs: List, fp: IO[bytes]):
        self.fp = None
        self._response = None
        self.msg = msg
        super().__init__(url, code, msg, hdrs, fp)

    @property
    def response(self):
        if not self._response and self.fp is not None:
            self._response = self.fp.read()
        return self._response

    @property
    def message(self) -> Union[str, None]:
        if self.code in [403, 400]:
            try:
                resp = json.loads(self.response)
                return resp.get('error', None) or resp.get('message', None)
            except JSONDecodeError:
                pass
        return None

    def __str__(self):
        msg = self.message
        if msg:
            return msg
        return str(super(HTTPError, self))


class PararamioRequestException(PararamioException):
    pass


class PararamioServerResponseException(PararamioRequestException):
    response: dict

    def __init__(self, msg: str, response: dict):
        self.msg = msg
        self.response = response

    def __str__(self):
        return f'{self.__class__.__name__}, {self.msg or " has been raised"}'


class PararamioLimitExceededException(PararamioRequestException):
    pass


class PararamioMethodNotAllowed(PararamioException):
    pass


class PararamioAuthenticationException(PararamioException):
    pass


class PararamioXSFRRequestError(PararamioAuthenticationException):
    pass


class PararamioPasswordAuthenticationException(PararamioAuthenticationException):
    pass


class PararamioSecondFactorAuthenticationException(PararamioAuthenticationException):
    pass


class PararamioCaptchaAuthenticationException(PararamioAuthenticationException):
    pass



import sys
from http.cookiejar import CookieJar
from typing import Any, Callable, Dict, Optional, Tuple, TypeVar, Union


if sys.version_info >= (3, 8):
    from typing import TypedDict


    class ProfileTypeT(TypedDict):
        unique_name: str
        id: int
        info: Optional[str]
        find_strict: bool
        name: str
        is_google: bool
        two_step_enabled: bool
        has_password: bool
        phoneconfirmed: bool
        email: str
        phonenumber: Optional[str]


    class PostMetaUserT(TypedDict):
        id: int
        name: str
        unique_name: str


    class PostMetaThreadT(TypedDict):
        title: str


    class PostMetaT(TypedDict):
        user: PostMetaUserT
        thread: PostMetaThreadT
else:
    ProfileTypeT = Dict[str, Any]
    PostMetaT = Dict[str, Any]

CookieJarT = TypeVar('CookieJarT', bound=CookieJar)
QuoteRangeT = Dict[str, Union[str, int]]
HeaderLikeT = Dict[str, str]
SecondStepFnT = Callable[[CookieJar, Dict[str, str], str], Tuple[bool, Dict[str, str]]]



import logging
import mimetypes
import os
from http.cookiejar import CookieJar, FileCookieJar, LoadError, MozillaCookieJar
from io import BytesIO
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

from pararamio._types import ProfileTypeT, SecondStepFnT
from pararamio.chat import Chat
from pararamio.constants import XSRF_HEADER_NAME
from pararamio.exceptions import PararamioAuthenticationException, PararamioHTTPRequestException, PararamioServerResponseException, PararamioValidationException
from pararamio.file import File
from pararamio.group import Group
from pararamio.post import Post
from pararamio.user import User
from pararamio.utils.authentication import authenticate, do_second_step, do_second_step_with_code, get_xsrf_token
from pararamio.utils.helpers import check_login_opts, get_empty_vars, lazy_loader
from pararamio.utils.requests import api_request, delete_file, download_file, upload_file, xupload_file


__all__ = ('Pararamio',)
log = logging.getLogger('pararamio.client')


class Pararamio:
    _login: Optional[str]
    _password: Optional[str]
    _key: Optional[str]
    _authenticated: bool
    _cookie: Union[CookieJar, FileCookieJar]
    __profile: Optional[dict]
    __headers: Dict[str, str]
    __user: dict

    def __init__(self, login: str = None, password: str = None, key: str = None, cookie: CookieJar = None, cookie_path: str = None,
                 ignore_broken_cookie: bool = False):
        self._login = login
        self._password = password
        self._key = key
        self.__headers = {}
        self.__profile = None
        self.__user = {}
        self._authenticated = False
        if cookie is not None:
            self._cookie = cookie
        elif cookie_path is not None:
            self._cookie = MozillaCookieJar(cookie_path)
            if os.path.exists(cookie_path):
                if not os.path.isfile(cookie_path):
                    raise OSError(f'path {cookie_path} is directory')
                if not os.access(cookie_path, os.R_OK):
                    raise OSError(f'file {cookie_path} is not readable')
                if not os.access(cookie_path, os.W_OK):
                    raise OSError(f'file {cookie_path} is not writable')
                try:
                    self._cookie.load(ignore_discard=True)
                    self._authenticated = True
                except LoadError as e:
                    log.error('failed to load cookie from file %s', cookie_path)
                    if not ignore_broken_cookie:
                        raise OSError(e)
        else:
            self._cookie = CookieJar()
        for cj in self._cookie:
            if cj.name == '_xsrf':
                self.__headers[XSRF_HEADER_NAME] = str(cj.value)
                break

    def _save_cookie(self) -> None:
        if isinstance(self._cookie, FileCookieJar):
            self._cookie.save(ignore_discard=True)

    def _profile(self, raise_on_error: bool = False) -> 'ProfileTypeT':
        return self.api_get('/user/me', raise_on_error=raise_on_error)

    def _do_auth(self, login: str, password: str, cookie_jar: CookieJar, headers: Dict[str, str], second_step_fn: SecondStepFnT,
                 second_step_arg: str, ) -> None:
        self._authenticated, user, xsrf = authenticate(login, password, cookie_jar, headers, second_step_fn, second_step_arg)
        if self._authenticated:
            self.__user = user
            self.__headers[XSRF_HEADER_NAME] = xsrf
            self._save_cookie()

    def _authenticate(self, login: str, password: str, second_step_fn: SecondStepFnT, second_step_arg: str) -> bool:
        if not self._cookie:
            self._do_auth(login, password, self._cookie, self.__headers, second_step_fn, second_step_arg)
        try:
            self._authenticated = True
            self._profile(raise_on_error=True)
        except PararamioHTTPRequestException:
            self._authenticated = False
            self._do_auth(login, password, self._cookie, self.__headers, second_step_fn, second_step_arg)
        return self._authenticated

    def authenticate(self, login: str = None, password: str = None, key: str = None) -> bool:
        """
        Authenticate client with totp key
        :param login: pararam login
        :param password: pararam password
        :param key: 16 chars second factor key to generate one time code
        :return: True if authentication success
        """
        login = login or self._login
        password = password or self._password
        key = key or self._key
        if not check_login_opts(login, password, key=key):
            raise PararamioAuthenticationException(f'{get_empty_vars(login=login, password=password, key=key)} must be set and not empty')
        return self._authenticate(login, password, do_second_step, key)  # type: ignore

    def authenticate_with_code(self, code: str, login: str = None, password: str = None) -> bool:
        """
        Authenticate client with generated TFA code
        :param login: pararam login
        :param password: pararam password
        :param code: 6 digits code
        :return: True if authentication success
        """
        login = login or self._login
        password = password or self._password
        if not check_login_opts(login, password, code=code):
            raise PararamioAuthenticationException(f'{get_empty_vars(login=login, password=password, code=code)} must be set and not empty')
        return self._authenticate(login, password, do_second_step_with_code, code)  # type: ignore

    def _api_request(self, url: str, method: str = 'GET', data: dict = None, callback: Callable = lambda rsp: rsp, raise_on_error: bool = False) -> Any:
        if not self._authenticated:
            self.authenticate()
        if not self.__headers.get(XSRF_HEADER_NAME, None):
            self.__headers[XSRF_HEADER_NAME] = get_xsrf_token(self._cookie)
            self._save_cookie()
        try:
            return callback(api_request(url, method, data, cookie_jar=self._cookie, headers=self.__headers))
        except PararamioHTTPRequestException as e:
            if raise_on_error:
                raise
            if e.code == 401:
                self._authenticated = False
                return self._api_request(url=url, method=method, data=data, callback=callback, raise_on_error=True)
            message = e.message
            if message == 'xsrf':
                log.info('xsrf is expire, invalid or was not set, trying to get new one')
                self.__headers[XSRF_HEADER_NAME] = ''
                return self._api_request(url=url, method=method, data=data, callback=callback, raise_on_error=True)
            raise

    def api_get(self, url: str, raise_on_error: bool = False) -> dict:
        return self._api_request(url, raise_on_error=raise_on_error)

    def api_post(self, url: str, data: Dict[Any, Any] = None, raise_on_error: bool = False) -> dict:
        return self._api_request(url, method='POST', data=data, raise_on_error=raise_on_error)

    def api_put(self, url: str, data: Dict[Any, Any] = None, raise_on_error: bool = False) -> dict:
        return self._api_request(url, method='PUT', data=data, raise_on_error=raise_on_error)

    def api_delete(self, url: str, data: Dict[Any, Any] = None, raise_on_error: bool = False) -> dict:
        return self._api_request(url, method='DELETE', data=data, raise_on_error=raise_on_error)

    @staticmethod
    def _get_file_info(file_path: str, filename: str = None, guess_type: bool = False) -> dict:
        if not os.path.exists(file_path):
            raise OSError(f'file {file_path} file does not exist')
        if not os.path.isfile(file_path):
            raise OSError(f'file {file_path} is a directory')
        if not os.access(file_path, os.R_OK):
            raise OSError(f'no read access to the {file_path} file')
        if filename is None:
            filename = os.path.basename(file_path)
        type_ = None
        if guess_type:
            if not mimetypes.inited:
                mimetypes.init(files=os.environ.get('PARARAMIO_MIME_TYPES_PATH', None))
            type_ = mimetypes.guess_type(os.path.basename(file_path))[0]
        return {
            'filename': filename,
            'size':     os.stat(file_path).st_size,
            'type':     type_
        }

    def _upload_file(self,
                     file_path: str,
                     chat_id: int,
                     filename: str = None,
                     type_: str = None,
                     organization_id: int = None,
                     reply_no: int = None,
                     quote_range: str = None):
        if not self._authenticated:
            self.authenticate()
        if not self.__headers.get(XSRF_HEADER_NAME, None):
            self.__headers[XSRF_HEADER_NAME] = get_xsrf_token(self._cookie)
        if type_ == 'organization_avatar' and organization_id is None:
            raise PararamioValidationException('organization_id must be set when type is organization_avatar')
        if type_ == 'chat_avatar' and chat_id is None:
            raise PararamioValidationException('chat_id must be set when type is chat_avatar')
        file_info = self._get_file_info(file_path, filename)
        fields = [
            ('type', type_),
            ('filename', file_info['filename']),
            ('size', file_info['size']),
            ('chat_id', chat_id),
            ('organization_id', organization_id),
            ('reply_no', reply_no),
            ('quote_range', quote_range)
        ]
        with open(file_path, 'rb') as f:
            return xupload_file(fp=f, fields=fields, headers=self.__headers, cookie_jar=self._cookie), dict(fields)

    def direct_upload_file(self, file_path: str, chat_id: int, filename: str = None, reply_no: int = None, quote_range: str = None) -> File:

        res, extra = self._upload_file(file_path=file_path, chat_id=chat_id, filename=filename, reply_no=reply_no, quote_range=quote_range)
        return File(self, guid=res['guid'], mime_type=extra['type'], **extra)

    def upload_file(self, file_path: str, chat_id: int, filename: str = None, is_avatar=False) -> File:
        res, extra = self._upload_file(file_path=file_path, chat_id=chat_id, filename=filename)
        return File(self, guid=res['guid'], mime_type=extra['type'], **extra)

    def old_upload_file(self, file_path: str, chat_id: int, filename: str = None, is_avatar=False) -> File:
        # not work
        perm_url = '/file/upload'
        file_info = self._get_file_info(file_path, filename, guess_type=True)
        perm_resp = self.api_post(perm_url, {
            'is_avatar': is_avatar,
            'thread_id': chat_id,
            'file':      {
                'name': file_info['filename'],
                'type': file_info['type'],
                'size': file_info['size']
            }
        })
        if 'perm' not in perm_resp:
            raise PararamioServerResponseException('invalid server permission response %s', perm_resp)
        perm = perm_resp['perm']
        with open(file_path, 'rb') as f:
            resp = upload_file(f, perm, filename=file_info['filename'], file_type=file_info['type'], headers=self.__headers, cookie_jar=self._cookie)
        if 'data' not in resp:
            log.error('no data in response %s', resp)
            raise PararamioServerResponseException('invalid server response exception %s', resp)
        return File(self, **resp['data'])

    def delete_file(self, guid: str) -> dict:
        return delete_file(guid, headers=self.__headers, cookie_jar=self._cookie)

    def download_file(self, guid: str, filename: str) -> BytesIO:
        return download_file(guid, filename, headers=self.__headers, cookie_jar=self._cookie)

    @property
    def profile(self) -> dict:
        if not self.__profile:
            self.__profile = self._profile()
        return self.__profile

    def search_user(self, query: str) -> List[User]:
        return User.search(self, query)

    def search_group(self, query: str) -> List[Group]:
        return Group.search(self, query)

    def search_posts(self, query: str, order_type: str = 'time', page: int = 1, chat_id: int = None, limit: Optional[int] = None) -> Tuple[int, Iterable[Post]]:
        return Chat.post_search(self, query, order_type=order_type, page=page, chat_id=chat_id, limit=limit)

    def list_chats(self) -> Iterable:
        url = '/core/chat/sync'
        chats_per_load = 50
        ids = self.api_get(url).get('chats', [])
        return lazy_loader(self, ids, Chat.load_chats, per_load=chats_per_load)

    def post_private_message_by_user_email(self, email: str, text: str) -> Post:
        url = '/msg/post/private'
        resp = self._api_request(url, method='POST', data={'text': text, 'user_email': email})
        return Post(Chat(self, resp['chat_id']), resp['post_no'])

    def post_private_message_by_user_id(self, user_id: int, text: str) -> Post:
        url = '/msg/post/private'
        resp = self._api_request(url, method='POST', data={'text': text, 'user_id': user_id})
        return Post(Chat(self, resp['chat_id']), resp['post_no'])

    def post_private_message_by_user_unique_name(self, unique_name: str, text: str) -> Post:
        url = '/msg/post/private'
        resp = self._api_request(url, method='POST', data={'text': text, 'user_unique_name': unique_name})
        return Post(Chat(self, resp['chat_id']), resp['post_no'])



import math
import uuid
import hashlib
from datetime import datetime, timedelta, timezone
from typing import Any, Callable, Dict, Iterable, List, Optional, Sequence, Tuple, cast

from pararamio.constants import DATETIME_FORMAT
from pararamio.exceptions import PararamioValidationException


__all__ = ('encode_digit', 'lazy_loader', 'encode_chat_id', 'join_ids', 'parse_datetime', 'check_login_opts', 'get_empty_vars',)


def check_login_opts(login: Optional[str], password: Optional[str], **kwargs: Optional[str]) -> bool:
    return all(map(bool, [login, password, kwargs.get('code', None) or kwargs.get('key', None)]))


def get_empty_vars(**kwargs: Any):
    return ', '.join([k for k, v in kwargs.items() if not v])


def encode_digit(digit: int, res: str = '') -> str:
    if not isinstance(digit, int):
        digit = int(digit)
    code_string = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_.'
    result = math.floor(digit / len(code_string))
    res = code_string[int(digit % len(code_string))] + res
    return encode_digit(result, res) if result > 0 else res


def encode_chat_id(chat_id: int, posts_count: int, last_read_post_no: int) -> str:
    return '-'.join(map(str, [chat_id, posts_count, last_read_post_no]))


def encode_chats_ids(chats_ids: List[Tuple[int, int, int]]) -> str:
    return '/'.join(map(lambda *args: encode_chat_id(*args), chats_ids))


def lazy_loader(cls: Any, items: Sequence, load_fn: Callable[[Any, List], List], per_load: int = 50) -> Iterable:
    load_counter = 0
    loaded_items: List[Any] = []
    counter = 0

    def load_items():
        return load_fn(cls, items[(per_load * load_counter): (per_load * load_counter) + per_load])

    for _ in items:
        if not loaded_items:
            loaded_items = load_items()
        if counter >= per_load:
            counter = 0
            load_counter += 1
            loaded_items = load_items()
        yield loaded_items[counter]
        counter += 1


def join_ids(items: List[Any]) -> str:
    return ','.join(map(str, items))


def get_utc(date: datetime) -> datetime:
    if date.tzinfo is None:
        raise PararamioValidationException('is not offset-aware datetime')
    return cast(datetime, date - cast(timedelta, date.utcoffset()))


def parse_datetime(data: Dict[str, Any], key: str, format_: str = DATETIME_FORMAT) -> Optional[datetime]:
    if key not in data:
        return None
    return datetime.strptime(data[key], format_).replace(tzinfo=timezone.utc)


def format_datetime(date: datetime) -> str:
    return get_utc(date).strftime(DATETIME_FORMAT)


def rand_id():
    _hash = hashlib.md5(bytes(uuid.uuid4().hex, 'utf8'))
    return str(int(int(_hash.hexdigest(), 16) * 10 ** -21))



from .chat import Chat
from .client import Pararamio
from .deferred_post import DeferredPost
from .group import Group
from .post import Post
from .team import Team, TeamMember
from .user import User
from .bot import PararamioBot
from .activity import Activity, ActivityAction


__all__ = (
    'Pararamio',
    'Chat',
    'Post',
    'DeferredPost',
    'User',
    'Group',
    'Team',
    'TeamMember',
    'PararamioBot',
    'ActivityAction',
    'Activity'
)



import os
import string
from os import path
from setuptools import find_packages, setup


def get_version(version_list):
    return '.'.join(map(str, version_list))


init = os.path.join(os.path.dirname(__file__), 'pararamio', 'constants.py')
version_line = list(filter(lambda l: l.startswith('VERSION'), open(init)))[0]
VERSION = get_version(''.join(list(filter(lambda c: c in string.digits + '.', version_line.split('=')[-1]))).split('.'))
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()
setup(
    name='pararamio',
    version=VERSION,
    description='Pararam Library',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/pararam-public/py-pararamio',
    author='Ilya Volnistov',
    author_email='i.volnistov@gaijin.team',
    license='MIT',
    packages=find_packages(),
    python_requires='>=3.6',
    zip_safe=False
)



from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union, Callable

from pararamio.chat import Chat
from pararamio.exceptions import PararamioRequestException
from pararamio.file import File
from pararamio.post import Post
from pararamio.utils.helpers import join_ids
from pararamio.activity import Activity, ActivityAction


if TYPE_CHECKING:
    from pararamio.client import Pararamio
    from datetime import datetime

__all__ = ('User',)


class User:
    __slots__ = ('_data', '_client', 'id', 'load_on_key_error',)
    id: int
    info: Optional[str]
    organizations: List[int]
    unique_name: str
    info_about_user: Optional[List[Dict[str, Any]]]
    name: str
    deleted: bool
    info_about_user_parsed: Optional[str]
    active: bool
    name_trans: str
    info_parsed: Optional[List[Dict[str, Any]]]
    _data: Dict[str, Any]
    _client: 'Pararamio'
    load_on_key_error: bool

    def __init__(self, client, id: int, load_on_key_error: bool = True, **kwargs):
        self._client = client
        self.id = id
        self._data = {'id': id, **kwargs}
        self.load_on_key_error = load_on_key_error

    def __getattr__(self, name):
        try:
            return self._data[name]
        except KeyError:
            if self.load_on_key_error:
                self.load()
                return self._data[name]
            raise

    def __eq__(self, other):
        if not isinstance(other, User):
            return id(other) == id(self)
        return self.id == other.id

    @property
    def has_pm(self):
        if hasattr(self, 'pm_thread_id') and self.pm_thread_id is not None:
            return True
        return False

    def get_pm_thread(self) -> Chat:
        pm_thread_id = self._data.get('pm_thread_id', None)
        if pm_thread_id is not None:
            chat = Chat(self._client, pm_thread_id)
            return chat
        return Chat.create_private_chat(self._client, self.id)

    def load(self) -> 'User':
        res = self._load_users_request(self._client, [self.id])
        if len(res) != 1:
            raise PararamioRequestException()
        self._data = {'id': self.id, **res[0]}
        return self

    @classmethod
    def _load_users_request(cls, client: 'Pararamio', ids: List[int]) -> dict:
        url = f'/user?ids={join_ids(ids)}'
        return client.api_get(url).get('users', [])

    @classmethod
    def load_users(cls, client, ids: List[int]) -> List['User']:
        return [cls(client, **user) for user in cls._load_users_request(client, ids)]

    def post(self, text: str, quote_range: Dict[str, Union[str, int]] = None, reply_no: int = None, file: File = None) -> Post:
        chat = self.get_pm_thread()
        return chat.post(text=text, quote_range=quote_range, reply_no=reply_no, file=file)

    def __str__(self):
        if 'name' not in self._data:
            self.load()
        return self._data.get('name')

    @classmethod
    def search(cls, client: 'Pararamio', search_string: str) -> List['User']:
        url = f'/users?flt={search_string}'
        return [cls(client, **user) for user in client.api_get(url).get('users', [])]

    def _activity_page_loader(self) -> Callable[..., Dict[str, Any]]:
        def loader(action: Optional[ActivityAction] = None, page: int = 1) -> Dict[str, Any]:
            action_ = action.value if action else ''
            url = f'/activity?user_id={self.id}&action={action_}&page={page}'
            return self._client.api_get(url)

        return loader

    def get_activity(self, start: 'datetime', end: 'datetime', actions: List[ActivityAction] = None) -> List[Activity]:
        """get user activity

        :param start: start time
        :param end: end time
        :param actions: list of action types (all actions if None)
        :returns: activity list
        """
        return Activity.get_activity(self._activity_page_loader(), start, end, actions)



from typing import Any, Callable, Dict, Iterable, List, Optional, Union, TYPE_CHECKING

from pararamio.exceptions import PararamioRequestException
from pararamio.utils import bot_request, join_ids, lazy_loader
from pararamio.activity import Activity, ActivityAction

if TYPE_CHECKING:
    from datetime import datetime


__all__ = ('PararamioBot',)


def _load_chats(cls, ids: List[int]) -> List[Dict[str, Any]]:
    url = f'/core/chat?ids={join_ids(ids)}'
    res = cls.request(url)
    if res and 'chats' in res:
        return [data for data in cls.request(url).get('chats', [])]
    raise PararamioRequestException('failed to load data for chats ids: {}'.format(','.join(map(str, ids))))


def _one_or_value_error(fn: Callable, msg: str, *args) -> Any:
    try:
        return fn()[0]
    except IndexError:
        pass
    raise ValueError(msg.format(*args))


class PararamioBot:
    key: str

    def __init__(self, key: str):
        if len(key) > 50:
            key = key[20:]
        self.key = key

    def request(self, url: str, method: str = 'GET', data: Optional[dict] = None, headers: dict = None) -> Dict:
        return bot_request(url, self.key, method=method, data=data, headers=headers)

    def profile(self):
        """

        :return: {
            "id":12981,
            "active":true,
            "deleted":false,
            "email":null,
            "find_strict":false,
            "has_password":false,
            "info":null,
            "info_parsed":[],
            "info_chat":43611,
            "is_bot":true,
            "is_google":false,
            "name":"ChatHelpBot",
            "name_trans":"",
            "unique_name":"helpbot",
            "organizations":[293,1],
            "phoneconfirmed":false,
            "phonenumber":null,
            "time_created":"2019-08-14T05:25:04.088435Z",
            "time_updated":"2019-09-04T04:37:26.065742Z",
            "two_step_enabled":false
        }
        """
        url = '/user/me'
        return self.request(url)

    def post_message(self, chat_id: int, text: str, reply_no: Optional[int] = None) -> Dict[str, Union[str, int]]:
        url = '/bot/message'
        return self.request(url, method='POST', data={'chat_id': chat_id, 'text': text, 'reply_no': reply_no})

    def post_private_message_by_user_id(self, user_id: int, text: str) -> Dict[str, Union[str, int]]:
        url = '/msg/post/private'
        return self.request(url, method='POST', data={'text': text, 'user_id': user_id})

    def post_private_message_by_user_email(self, email: str, text: str) -> Dict[str, Union[str, int]]:
        url = '/msg/post/private'
        return self.request(url, method='POST', data={'text': text, 'user_email': email})

    def post_private_message_by_user_unique_name(self, unique_name: str, text: str) -> Dict[str, Union[str, int]]:
        url = '/msg/post/private'
        return self.request(url, method='POST', data={'text': text, 'user_unique_name': unique_name})

    def get_tasks(self) -> Dict[str, Any]:
        url = '/msg/task'
        return self.request(url)

    def set_task_status(self, chat_id: int, post_no: int, state: str) -> Dict:
        if str.lower(state) not in ('open', 'done', 'close'):
            raise ValueError(f'unknown state {state}')
        url = f'/msg/task/{chat_id}/{post_no}'
        data = {'state': state}
        return self.request(url, method='POST', data=data)

    def get_chat(self, chat_id) -> Dict[str, Any]:
        url = f'/core/chat?ids={chat_id}'
        return _one_or_value_error(lambda: self.request(url).get('chats', []), 'chat with id {0} is not found', chat_id)

    def get_chats(self) -> Iterable[dict]:
        url = '/core/chat/sync'
        chats_per_load = 50
        ids = self.request(url).get('chats', [])
        return lazy_loader(self, ids, _load_chats, per_load=chats_per_load)

    def get_users(self, users_ids: List[int]) -> Dict[str, Any]:
        url = f'/core/user?ids={join_ids(users_ids)}'
        return self.request(url).get('users', [])

    def get_user_by_id(self, user_id: int):
        return _one_or_value_error(lambda: self.get_users([user_id]), 'user with id {0} is not found', user_id)

    def _user_activity_page_loader(self, user_id: int) -> Callable[..., Dict[str, Any]]:
        def loader(action: ActivityAction = None, page: int = 1) -> Dict[str, Any]:
            action_ = action.value if action else ''
            url = f'/activity?user_id={user_id}&action={action_}&page={page}'
            return self.request(url)

        return loader

    def get_user_activity(self, user_id: int, start: 'datetime', end: 'datetime', actions: List[ActivityAction] = None) -> List[Activity]:
        """get user activity by user_id

        :param user_id: user id
        :param start: start time
        :param end: end time
        :param actions: list of action types (all actions if None)
        :returns: activity list
        """
        return Activity.get_activity(self._user_activity_page_loader(user_id), start, end, actions)
